from veem.models.base import Base
from veem.models.amount import Amount
from veem.models.account import Account
from veem.models.attachment import Attachment
from veem.models.exchange_rate import ExchangeRate
from veem.models.payment_approval import PaymentApproval
from veem.models.push_payment_info import PushPaymentInfo

from veem.utils import deseralize
from veem.constants import PaymentStatus

class Payment(Base):
    def __init__(self,
                 id=None,
                 status=None,
                 exchangeRate=None,
                 timeCreated=None,
                 claimLink=None,
                 pushPaymentInfo=None,
                 paymentApproval=None,
                 batchItemId=None,
                 payee=None,
                 payer=None,
                 clientId=None,
                 amount=None,
                 notes=None,
                 externalInvoiceRefId=None,
                 ccEmails=[],
                 purposeOfPayment=None,
                 attachments=[],
                 exchangeRateQuoteId=None,
                 **kwargs):

        self._validate_constants(PaymentStatus, status)

        self.id = id
        self.status = status
        self.exchangeRate = deseralize(ExchangeRate, exchangeRate)
        self.timeCreated = timeCreated
        self.claimLink = claimLink
        self.pushPaymentInfo = deseralize(PushPaymentInfo, pushPaymentInfo)
        self.paymentApproval = deseralize(PaymentApproval, paymentApproval)
        self.batchItemId = batchItemId
        self.payee = deseralize(Account, payee)
        self.payer = deseralize(Account, payer)
        self.clientId = clientId
        self.amount = deseralize(Amount, amount)
        self.notes = notes
        self.externalInvoiceRefId = externalInvoiceRefId
        self.ccEmails = ccEmails
        self.purposeOfPayment = purposeOfPayment
        self.attachments = [deseralize(Attachment,
                                       attach) for attach in attachments]
        self.exchangeRateQuoteId = exchangeRateQuoteId

    def _generate_request(self, approveAutomatically=False):
        return PaymentRequest(
                     batchItemId=self.batchItemId,
                     exchangeRateQuoteId=self.exchangeRateQuoteId,
                     payee=self.payee,
                     amount=self.amount,
                     notes=self.notes,
                     externalInvoiceRefId=self.externalInvoiceRefId,
                     ccEmails=self.ccEmails,
                     purposeOfPayment=self.purposeOfPayment,
                     attachments=self.attachments,
                     approveAutomatically=approveAutomatically,
                )




from veem.models.base import Base

class ContactListParameters(Base):
    def __init__(self,
                email=None,
                firstName=None,
                lastName=None,
                businessName=None,
                batchId=None,
                batchItemIds=None,
                pageNumber=0,
                pageSize=1,
                 **kwargs):

        self.email = email
        self.firstName = firstName
        self.lastName = lastName
        self.businessName = businessName
        self.batchId = batchId
        self.batchItemIds = batchItemIds
        self.pageNumber = pageNumber
        self.pageSize = pageSize



from veem.models.base import Base

class ExchangeRateRequest(Base):
    def __init__(self,
                 fromAmount=None,
                 fromCurrency=None,
                 fromCountry=None,
                 toAmount=None,
                 toCurrency=None,
                 toCountry=None,
                 **kwargs):

        self.fromAmount = fromAmount
        self.fromCurrency = fromCurrency
        self.fromCountry = fromCountry
        self.toAmount = toAmount
        self.toCurrency = toCurrency
        self.toCountry = toCountry




from veem.models.base import Base

class PurposeOfPayment(Base):
    def __init__(self,
                 purposeCode=None,
                 description=None,
                 **kwargs):

        self.purposeCode = purposeCode
        self.description = description




from veem.models.base import Base

class ExchangeRate(Base):
    def __init__(self,
                 hashId=None,
                 fromAmount=None,
                 fromCurrency=None,
                 fromCountry=None,
                 toAmount=None,
                 toCurrency=None,
                 toCountry=None,
                 rate=None,
                 expiry=None,
                 **kwargs):

        self._validate_currency_code(fromCurrency)
        self._validate_currency_code(toCurrency)

        self._validate_country_code(fromCountry)
        self._validate_country_code(toCountry)

        self.fromAmount = fromAmount
        self.fromCurrency = fromCurrency
        self.fromCountry = fromCountry
        self.toAmount = toAmount
        self.toCurrency = toCurrency
        self.toCountry = toCountry
        self.hashId = hashId
        self.rate = rate
        self.expiry = expiry

    def _generate_request(self):
        return ExchangeRateRequest(
                     fromAmount=self.fromAmount,
                     fromCurrency=self.fromCurrency,
                     fromCountry=self.fromCountry,
                     toAmount=self.toAmount,
                     toCurrency=self.toCurrency,
                     toCountry=self.toCountry,
                )




from veem.models.base import Base

class VeemErrorResponse(Base):
    def __init__(self,
                 message=None,
                 code=None,
                 logTag=None,
                 timestamp=None,
                 fileName=None,
                 lineNumber=None,
                 **kwargs):

        self.message = message
        self.code = code
        self.logTag = logTag
        self.timestamp = timestamp
        self.fileName = fileName
        self.lineNumber = lineNumber



from veem.models.base import Base
from veem.models.account import Account


class AccountResponse(Base):
    def __init__(self,
                 id=None,
                 firstName=None,
                 lastName=None,
                 name=None,
                 email=None,
                 isoCountryCode=None,
                 isContact=None,
                 **kwargs):

        self.id = id
        self.firstName = firstName
        self.lastName = lastName
        self.name = name
        self.email = email
        self.isoCountryCode = isoCountryCode
        self.isContact = isContact

    @property
    def convert(self):
        return Account(**self.json)




class VeemException(Exception):

    def __init__(self, errorResponse=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.errorCode = getattr(errorResponse, 'errorCode', None)
        self.message = getattr(errorResponse, 'message', None)
        self.timestamp = getattr(errorResponse, 'timestamp', None)
        self.logTag = getattr(errorResponse, 'logTag', None)
        self.fileName = getattr(errorResponse, 'fileName', None)
        self.lineNumber = getattr(errorResponse, 'lineNumber', None)
        if isinstance(errorResponse, dict):
            for key in ('errorCode', 'message', 'timestamp', 'logTag',
                        'fileName', 'lineNumber'):
                setattr(self, key,
                        errorResponse.get(key, getattr(self, key, None)))

class VeemBadRequestException(VeemException):
    pass

class VeemConflictException(VeemException):
    pass

class VeemForbiddenException(VeemException):
    pass

class VeemInternalException(VeemException):
    pass

class VeemNotFoundException(VeemException):
    pass

class VeemConflictException(VeemException):
    pass

class VeemSdkException(VeemException):
    pass

class VeemUnauthorizedException(VeemException):
    pass



# import from python libraries
import uuid
import logging
import requests
from functools import partial
from urllib.parse import urljoin
from requests.auth import HTTPBasicAuth

from veem import __API_VERSION__ as API_VERSION
from veem.exceptions import VeemException

logger = logging.getLogger(__name__)

class ApiError(Exception):
    ''' ApiError

    The abstract class for ApiError

    '''
    def __init__(self, request, response):
        super(ApiError, self).__init__()
        self.request = request
        self.response = response

class ApiRequestError(ApiError):
    pass

class ApiJsonError(ApiError):
    pass

class VeemRestApi(object):

    def __init__(self, url=None, session=None, endpoints=dict()):
        self.url = url if url.endswith('/') else '%s/' % url
        self.session = session or requests.Session()
        self.endpoints = endpoints

    def __getattr__(self, attr):
        """ dunder getattr method

        magic method that use endpoints list defined at ENDPOINTS and
        functools.partial to link class method

        """
        if attr in self.endpoints:
            return partial(self._api_call, attr)
        raise AttributeError("'%s' object has no attribute '%s'" % (
                                    self.__class__.__name__, attr))

    def _generate_endpoint(self, api_route, value, **kwargs):
        value = value.format(**kwargs)
        value = value if not api_route else 'veem/v{}/{}{}'.format(
            API_VERSION, api_route, value
        )
        return value.lstrip('/')

    def _api_call(self, attr,
                  uri_params={}, access_token=None, api_route=None,
                  request_header={}, request_params={}, request_auth=[],
                  request_files={}, request_body=None,
                  request_stream=False, *args, **kwargs):
        """ _api_call method

        main method that makes the REST api call

        """
        if attr not in self.endpoints:
            raise AttributeError("'%s' object has no suitable api endpoint" % (
                                        self.__class__.__name__))
        method, endpoint = self.endpoints[attr]
        endpoint = self._generate_endpoint(api_route, endpoint, **uri_params)
        method = (method or 'get').upper()
        url = urljoin(self.url, endpoint.lstrip('/'))
        try:
            if access_token:
                self.session.headers.update(
                    {'Authorization':'Bearer %s' % getattr(
                                access_token, 'accessToken', access_token),
                     'X-REQUEST-ID': str(uuid.uuid4())})

            auth = HTTPBasicAuth(*request_auth) if request_auth else None
            # send the REST api call
            request = requests.Request(method, url,
                                       json=request_body or kwargs,
                                       files=request_files,
                                       params=request_params,
                                       auth=auth)
            prepared = request.prepare()
            prepared.headers.update(self.session.headers)
            response = self.session.send(prepared)

            # if streaming request don't bother to validate jsonize the data
            if request_stream:
                return response

            # validate response
            if not response.ok:
                raise ApiRequestError(request, response)
            json = response.json()
            if not json:
                raise ApiJsonError(request, response)
        except ApiError as ae:
            try:
                json = response.json()
            except Exception as aee:
                json = {}
            # try to unpack response exception
            try:
                response = ae.response.text
            except Exception as aee:
                response = str(aee)
            logger.error("'%s' endpoint request failed: %s" % (attr, response))
            if json:
                raise VeemException(errorResponse=json)
            return response
        except Exception as e:
            logger.error("'%s' endpoint request errored: %s" % (attr, str(e)))
            return str(e)

        return json



from veem.models.base import Base

class PaymentApprovalResponse(Base):
    def __init__(self,
                 status=None,
                 approverNumber=None,
                 approverNumberRequired=None,
                 userApprovalResponseList=[],
                 **kwargs):

        self.status = status
        self.approverNumber = approverNumber
        self.approverNumberRequired = approverNumberRequired
        self.userApprovalResponseList = userApprovalResponseList



from .address import AddressRequest
from veem.models.base import Base

class BankAccountRequest(Base):
    def __init__(self,
                 routingNumber=None,
                 bankName=None,
                 bankAccountNumber=None,
                 currencyCode=None,
                 isoCountryCode=None,
                 iban=None,
                 swiftBic=None,
                 beneficiaryName=None,
                 bsbBankCode=None,
                 branchCode=None,
                 transitCode=None,
                 bankInstitutionNumber=None,
                 bankIfscBranchCode=None,
                 sortCode=None,
                 bankCode=None,
                 clabe=None,
                 bankCnaps=None,
                 bankAddress=None,
                 **kwargs):

        self.routingNumber = routingNumber
        self.bankName = bankName
        self.bankAccountNumber = bankAccountNumber
        self.currencyCode = currencyCode
        self.isoCountryCode = isoCountryCode
        self.iban = iban
        self.swiftBic = swiftBic
        self.beneficiaryName = beneficiaryName
        self.bsbBankCode = bsbBankCode
        self.branchCode = branchCode
        self.transitCode = transitCode
        self.bankInstitutionNumber = bankInstitutionNumber
        self.bankIfscBranchCode = bankIfscBranchCode
        self.sortCode = sortCode
        self.bankCode = bankCode
        self.clabe = clabe
        self.bankCnaps = bankCnaps
        self.bankAddress = bankAddress



from veem.models.base import Base

from veem.utils import deseralize

class Page(Base):
    def __init__(self,
                 number=None,
                 size=None,
                 totalPages=None,
                 numberOfElements=None,
                 totalElements=None,
                 previousPage=None,
                 nextPage=None,
                 first=None,
                 last=None,
                 content=[],
                 **kwargs):

        self.number = number
        self.size = size
        self.totalPages = totalPages
        self.numberOfElements = numberOfElements
        self.totalElements = totalElements
        self.previousPage = previousPage
        self.nextPage = nextPage
        self.first = first
        self.last = last
        self.content = content

    def convertContent(self, cls):
        self.content = [deseralize(cls, c) for c in self.content]
        return self




from veem.models.base import Base

from veem.constants import ApprovalStatus

class Approver(Base):
    def __init__(self,
                 status=None,
                 email=None,
                 firstName=None,
                 lastName=None,
                 middleName=None,
                 **kwargs):

        self._validate_constants(ApprovalStatus, status)

        self.status = status
        self.email = email
        self.firstName = firstName
        self.lastName = lastName
        self.middleName = middleName



from veem.models.base import Base

class AmountResponse(Base):
    def __init__(self,
                 number=None,
                 currency=None,
                 **kwargs):

        self.number = number
        self.currency = currency




import os
import shutil
import tempfile

from veem.constants import FILE_DOWNLOAD_CHUNK_SIZE
from veem.client.responses.attachment import AttachmentResponse

from veem.client.base import Base
from veem.utils.rest import VeemRestApi
from veem.utils import file_access_check, extract_file_content_type

from veem.models.attachment import Attachment
from veem.exceptions import VeemSdkException

class AttachmentClient(Base):

    def __init__(self, config, **kwargs):

        self.config = config
        self.context = config.context
        self.client = VeemRestApi(self.config.url,
                                  self.context.session,
                                  dict(upload=('post', ''),
                                       download=('get', '')))

    def upload(self, attachment, description):
        """
            Upload an attachment to Veem. The attachment can be referenced from
            other entities, like Payment and Invoice.

            @param attachment: An Attachment with a valid file path to a
                               readable file
            @param description: Attachment description
            @return A new Attachment with name and referenceId set
            @throws VeemException If the provided Attachment path is invalid, or
                                  if uploading fails.
        """
        attachment = Attachment(**attachment) if isinstance(
                                attachment, dict) else attachment
        file_access_check(attachment.path, raise_exception=True)
        content_type = extract_file_content_type(attachment.path)
        request_files = dict(file=(os.path.basename(attachment.path),
                                   open(attachment.path, 'rb'),
                                   content_type))
        return self._response_handler(
                        AttachmentResponse,
                        self.client.upload(access_token=self.context.token,
                                        api_route='attachments',
                                        description=description,
                                        request_files=request_files)
                            )

    def download(self, request,
                 targetDirectory=None,
                 chunk_size=FILE_DOWNLOAD_CHUNK_SIZE):
        """
            Download an attachment from Veem.

            @param attachment: An Attachment with a valid name and referenceId
            @param targetDirectory: The directory where the downloaded file will
                                    be stored
            @param chunk_size: The data chunk size to write each iteration
            @return target file path to the downloaded attachment
            @throws VeemException If the provided targetDirectory is invalid, or
                                  if downloading fails.
        """
        with self.client.download(access_token=self.context.token,
                                  api_route='attachments',
                                  request_stream=True,
                                  request_params=request.json) as stream:
            stream.raise_for_status()
            kwargs = dict(delete=False)
            # if target directory is provided, calculate path and check for
            # write access
            if targetDirectory:
                if not os.path.isdir(targetDirectory):
                    if not os.path.isfile(targetDirectory):
                        raise VeemSdkException("Unable to access directory")
                    targetDirectory = os.path.abspath(targetDirectory)

                if not os.access(targetDirectory, os.W_OK):
                    raise VeemSdkException("Target directory is not writable")

                kwargs['dir'] = targetDirectory
            # generate a temporay file, save the download content and return the
            # full path
            with tempfile.NamedTemporaryFile(**kwargs) as fp:
                for chunk in stream.iter_content(chunk_size=chunk_size):
                    if chunk:
                        fp.write(chunk)
                return fp.name
